#ifndef SMARTPTR_H
#define SMARTPTR_H
//========================================================================
// SmartPtr.h : Smart Pointer class
//
// Part of the GameCode Application
//
// GameCode is the sample application that encapsulates much of the source code
// discussed in "Game Coding Complete" by Mike McShaffry, published by
// Paraglyph Press. ISBN: 1-932111-75-1
//
// If this source code has found it's way to you, and you think it has helped you
// in any way, do the author a favor and buy a new copy of the book - there are 
// detailed explanations in it that compliment this code well. Buy a copy at Amazon.com
// by clicking here: http://www.amazon.com/exec/obidos/ASIN/1932111751/gamecodecompl-20/
//
// There's also a companion web site at http://www.mcshaffry.com/GameCode/portal.php
//
// Distributed with the permission of Sandu Turcan
//
//========================================================================

///////////////////////////////////////////////////////
//
//  Original Code by Sandu Turcan
//  Posted on the Code Project: http://www.codeproject.com/
//
//  Code modified from the original, to conform with 
//  coding standards in "Game Coding Complete"
//
///////////////////////////////////////////////////////

///////////////////////////////////////////////////////
//  
//  This code is discussed in Game Coding Complete
//  Chapter 3, page 54-59
//
///////////////////////////////////////////////////////

template <class T> class SmartPtr;

/////////////////////////////////////////
// IRefCount
// is an interface for reference counting
// Classes can implement it themselves,
// or SmartPtr will provide its internal implementation of IRefCount

template <class T> class IRefCount {
    friend class SmartPtr<T>;
    protected:
    virtual void AddRef() = 0;
    virtual void Release() = 0;
    virtual T * GetPtr() const = 0;
};

//============================================================
/////////////////////////////////////////
// IRefCountImpl
// is a standart implementation of IRefCount
// To use it just derive your class from it:
// class CMyObject : public IRefCountImpl<CMyObject> { ... };
// Reminder: implementing IRefCount is optional but it would reduce
// memory fragmentation. 

template <class T> class IRefCountImpl 
  : public IRefCount<T> 
{
private:
    int m_Count;

protected:
    virtual void AddRef() { m_Count++; }
    virtual void Release()
    {
		assert(m_Count>=0);
        m_Count--;
        if(m_Count<=0)
        {
            Destroy();
        }
    }
    virtual T * GetPtr() const { return ((T *)this); }
    virtual void Destroy() { if(GetPtr()!=NULL) delete GetPtr(); }

	IRefCountImpl() { m_Count = 0; } 
};

//============================================================

/////////////////////////////////////////
// SmartPtr

template <class T> class SmartPtr
{
private:
    IRefCount<T> *m_RefCount;
    
    /////////////////////////////////////////
    // RefCounter
    // An internal implementation of IRefCount
    // for classes that don't implement it
    // SmartPtr will automatically choose between its internal and 
    // class' implementation of IRefCount
    class RefCounter : public IRefCountImpl<T> {
        private:
        T *m_Ptr;

        protected:
        virtual T * GetPtr() const { return m_Ptr; }
        virtual void Destroy()	{ delete this; }

        public:
        RefCounter(T *ptr)		{ m_Ptr = ptr; }
        virtual ~RefCounter()	{ IRefCountImpl<T>::Destroy(); }
    };

    // this method is called if T does not implement IRefCount
    void Assign(void *ptr)
    {
        if(ptr==NULL)
            Assign((IRefCount<T> *)NULL);
        else
        {
            Assign(new RefCounter(static_cast<T *>(ptr)));
        }
    }
    // this method is picked over Assign(void *ptr)
    // if T implements IRefCount.
    // This allows some memory usage optimization
    void Assign(IRefCount<T> *refcount)
    {
        if( refcount!=NULL ) 
			refcount->AddRef();
        IRefCount<T> *oldref = m_RefCount;
        m_RefCount = refcount;
        if ( oldref!=NULL ) 
			oldref->Release();
    }

public:
    SmartPtr()						{ m_RefCount = NULL; }
    SmartPtr(T * ptr)				{ m_RefCount = NULL; Assign(ptr); }
    SmartPtr(const SmartPtr &sp)	{ m_RefCount = NULL; Assign(sp.m_RefCount); }
    virtual ~SmartPtr()				{ Assign((IRefCount<T> *)NULL); }

    // get the contained pointer, not really needed but...
	T *GetPtr() const		{ return (m_RefCount==NULL) ? NULL : m_RefCount->GetPtr(); }

    // assignment operators
    SmartPtr& operator = (const SmartPtr &sp) { Assign(sp.m_RefCount); return *this; }
    SmartPtr& operator = (T * ptr)			  { Assign(ptr); return *this; }

	// T access and const conversion
    T * operator ->()						{ assert(GetPtr()!=NULL); return GetPtr(); }
    operator T* () const					{ return GetPtr(); }

    // utilities
    bool operator !()						{ return GetPtr()==NULL; }
    bool operator ==(const SmartPtr &sp)	{ return GetPtr()==sp.GetPtr(); }
    bool operator !=(const SmartPtr &sp)	{ return GetPtr()!=sp.GetPtr(); }
	bool operator  <(const SmartPtr &sp)	{ return GetPtr()<sp.GetPtr(); }
};
#endif
